from __future__ import with_statement # requires Python 2.5
import os
import struct

import ssconst
import ssmisc
from ssutil import writeUint32BE, generate_sound_text

class TextResourceExporter(object):
    def export_to_resource(self, text_data, text_path, speech_map, id_to_game_offset_map):
        raise NotImplementException()

class TextMonsterExporter(TextResourceExporter):
    def export_to_resource(self, text_data, text_path, speech_map, id_to_game_offset_map):
        with file(text_path, 'w') as text_file:
            for i, line in enumerate(text_data):
                sounds = speech_map.get_sound_used_by_line(i)
                if len(sounds) > 0:
                    sound_num, s = sounds[0] # only support one sound per line
                    offset = id_to_game_offset_map[s.id]
                    aux_data_size = speech_map.get_aux_data_size_of_sound(sound_num)
                    sound_text = generate_sound_text(offset, aux_data_size)
                    line = sound_text + line
                # Not sure what's going on, we seem to lose some newlines
                # somewhere along the line, so add it back in (only if
                # the line is missing a trailing newline character).
                text_file.write(line + ('' if line.endswith("\n") else "\n"))

class TextBunExporter(TextResourceExporter):
    def export_to_resource(self, text_data, text_path, speech_map, id_to_game_offset_map):
        with file(text_path, 'w') as text_file:
            for i, line in enumerate(text_data):
                sounds = speech_map.get_sound_used_by_line(i)
                if len(sounds) > 0:
                    sound_num, s = sounds[0] # only support one sound per line
                    sound_id = "/" + id_to_game_offset_map[s.id] + "/"
                    line = sound_id + line
                # Not sure what's going on, we seem to lose some newlines
                # somewhere along the line, so add it back in (only if
                # the line is missing a trailing newline character).
                text_file.write(line + ('' if line.endswith("\n") else "\n"))

class SoundResourceExporter(object):
    def __init__(self):
        self.compression_type = None

    def export_to_resource(self, path):
        """Writes sound data to a file at the given path.

        Returns a mapping of game IDs to a value that can be used in
        the game to trigger sounds. For original resources, this will be the
        actuall offset of the sound file in the resource. For compressed
        resources, it will just be the ID."""
        pass

class MonsterExporter(SoundResourceExporter):
    resource_type = "MONSTER"

class OriginalMonsterExporter(MonsterExporter):
    pass

class CompressedMonsterExporter(MonsterExporter):
    def export_to_resource(self, path, speech_map):
        sounds = speech_map.sounds
        # Use this when writing the offset table
        id_to_offset_map = {}
        # Converts IDs to offsets used in the game. Not needed for compressed
        #  files (we can insert IDs as-is). Original resources need the
        #  actual offsets of the VOC files.
        id_to_game_offset_map = {}
        buffer_size = 1024000

        with file(path, 'wb') as res_file:
            # Write the size of the offset lookup table
            table_size = len(sounds) * struct.calcsize('>4I')
            writeUint32BE(res_file, table_size)

            # Write dummy offset lookup table
            offset_table_location = res_file.tell()
            res_file.write('\x00' * table_size)

            # Write the sound data (including metadata)
            for s in sounds:
                # Don't export unused sounds.
                # (commented out because it caused problems when writing the
                #  offset table)
                #if len(s.lines_used) == 0:
                #    continue

                # Keep track of real offset of sound data
                # (4 bytes is the size of the "table size" value)
                id_to_offset_map[s.id] = res_file.tell() - table_size - 4

                # Write lip sync tags (metadata)
                tags = s.metadata['LipSynch']
                tags = struct.pack('>' + str(len(tags)) + 'H', *tags)
                res_file.write(tags)

                # Map the sound ID to itself (don't need to put the real offset)
                id_to_game_offset_map[s.id] = int(s.id)

                # Open source sound file, copy data from there (buffered)
                with file(s.source.path, 'rb') as source_file:
                    left_to_read = s.source.size
                    source_file.seek(s.source.offset)
                    while left_to_read > 0:
                        read_size = min(left_to_read, buffer_size)
                        res_file.write(source_file.read(read_size))
                        left_to_read -= read_size
                
            # Write the offset table
            res_file.flush()
            res_file.seek(offset_table_location)
            for s in sounds:
                orig_off = int(s.id) # could be any value any value but must be int-able.
                new_off = id_to_offset_map[s.id] # must be real offset
                num_tags = len(s.metadata['LipSynch']) * 2 # list of 16-bit values
                sound_size = s.source.size # size of the sound data (minus tags)

                writeUint32BE(res_file, (orig_off, new_off, num_tags, sound_size))

        return id_to_game_offset_map

class CompressedBunExporter(SoundResourceExporter):
    compressed_types_to_ext = {
        ssconst.COMP_OGG : ".ogg",
        ssconst.COMP_MP3 : ".mp3",
        ssconst.COMP_FLAC : ".fla"
    }

    def _save_rmap(self, res_file, metadata):
        res_file.write("RMAP")
        writeUint32BE(
            res_file,
            (
                metadata["Version"],
                metadata["BitDepth"],
                metadata["Frequency"],
                metadata["Channels"],
                len(metadata["Regions"]),
                len(metadata["Jumps"]),
                len(metadata["Syncs"]),
                len(metadata["Markers"])
            )
        )
        assert len(metadata["Regions"]) <= 1, (
            "ScummSpeaks does not support sound files with more than 1 region.")
        for region in metadata["Regions"]:
            writeUint32BE(
                res_file,
                (   region["Offset"],
                    region["Length"]
                )
            )
        for jump in metadata["Jumps"]:
            writeUint32BE(
                res_file,
                (
                    jump["Offset"],
                    jump["Destination"],
                    jump["HookId"],
                    jump["FadeDelay"]
                )
            )
        for sync in metadata["Syncs"]:
            data = sync["SyncData"]
            writeUint32BE(
                res_file,
                (
                    len(data)

                )
            )
            res_file.write(data)
        for marker in metadata["Markers"]:
            data = marker["MarkerData"]
            writeUint32BE(
                res_file,
                (
                    marker["Position"],
                    len(data)
                )
            )
            res_file.write(data)

    def export_to_resource(self, path, speech_map):
        sounds = speech_map.sounds
        # Not really needed for BUN files, just for API compatibility.
        id_to_game_id_map = {}
        buffer_size = 1024000

        with file(path, 'wb') as res_file:
            res_file.write("LB23")
            # Write placeholder for table offset & number of entries.
            writeUint32BE(res_file, (0, 0))

            # Keep track of files written to the BUN, so we can store it
            #  later in the table at the end of the file.
            bun_file_list = []

            for s in sounds:
                # Don't export unused sounds.
                if len(s.lines_used) == 0:
                    continue
                
                # Save the RMAP from metadata
                rmap_name = s.id + ".map"
                rmap_name += "\x00" * (24 - len(rmap_name)) # pad with null chars
                rmap_offset = res_file.tell()
                self._save_rmap(res_file, s.metadata)
                rmap_size = res_file.tell() - rmap_offset

                bun_file_list.append(
                    (rmap_name, rmap_offset, rmap_size)
                )

                # Open source sound file, copy data from there (buffered)
                sound_name = (s.id +
                              "_reg000" +
                              self.compressed_types_to_ext[speech_map.compression_type])
                sound_name += "\x00" * (24 - len(sound_name)) # pad with null chars
                sound_offset = res_file.tell()
                with file(s.source.path, 'rb') as source_file:
                    left_to_read = s.source.size
                    source_file.seek(s.source.offset)
                    while left_to_read > 0:
                        read_size = min(left_to_read, buffer_size)
                        res_file.write(source_file.read(read_size))
                        left_to_read -= read_size
                sound_size = s.source.size

                bun_file_list.append(
                    (sound_name, sound_offset, sound_size)
                )

                # Dummy mapping of ID for use in game scripts.
                id_to_game_id_map[s.id] = s.id

            table_offset = res_file.tell()
            table_num_entries = len(bun_file_list)

            # Write the table data
            for name, offset, size in bun_file_list:
                assert len(name) == 24, (
                "Names for files contained in a BUN resource " +
                "need to be padded or restricted to 24 characters")
                res_file.write(name)
                writeUint32BE(res_file, (offset, size))
            res_file.flush()
            res_file.seek(4)
            writeUint32BE(res_file, (table_offset, table_num_entries))
        return id_to_game_id_map

class ResourceExporterFactory(object):
    compressed_types = frozenset([
        ssconst.COMP_OGG,
        ssconst.COMP_MP3,
        ssconst.COMP_FLAC
    ])
    def __new__(cls, speech_map, *args, **kwds):
        resource_type = speech_map.resource_type
        compression_type = speech_map.compression_type
        if resource_type == ssconst.TYPE_MONSTER:
            if compression_type in cls.compressed_types:
                return CompressedMonsterExporter()
            else:
                return None # TODO: support original MONSTER files
        elif resource_type == ssconst.TYPE_BUN:
            if compression_type in cls.compressed_types:
                return CompressedBunExporter() # those poor squashed buns...
            else:
                return None # TODO: support original BUN files
        else:
            raise ssmisc.SSException("Unknown resource type: %s" % resource_type)

class TextExporterFactory(object):
    def __new__(cls, speech_map, *args, **kwds):
        resource_type = speech_map.resource_type
        if resource_type == ssconst.TYPE_MONSTER:
            return TextMonsterExporter()
        elif resource_type == ssconst.TYPE_BUN:
            return TextBunExporter()
        else:
            raise ssmisc.SSException("Unknown resource type: %s" % resource_type)