from __future__ import with_statement # requires Python 2.5
import math
import os
import xml.etree.cElementTree as et

import mutagen.oggvorbis
import mutagen.mp3
import mutagen.flac

import ssconst
import ssmisc
import ssutil

class SoundSource(object):
    def __init__(self, path=None, offset=None, size=None):
        """
        ? type = "MONSTER", "BUN", "file"
        path = path to source file
        offset = offset within the source file to the sound
        size = size of the sound data
        """
        self.path = path
        self.offset = offset
        self.size = size

    def __str__(self):
        s = (
            ("SoundSource: [path: %s, offset: %s, size: %s]")
            % (self.path, self.offset, self.size)
        )
        return s

class SoundEntry(object):
    def __init__(self, id=None, source=None, metadata=None, lines_used=None):
        """
        id = the (original) offset used in-game.
        source = SoundSource object
        metadata = dict of metadata (lip synch tags, RMAP data for BUNs)
        lines_used = list of ints, representing the line numbers that use this sound.
        """
        self.id = id
        self.source = source
        self.metadata = metadata
        self.lines_used = lines_used

    def __str__(self):
        s = (
            ("SoundEntry: [id: %s, source: %s, metadata: %s, lines_used: %s]")
            % (self.id, self.source, self.metadata, self.lines_used)
        )
        return s

class SpeechMap(object):
    """
    Stores the mapping of sound files to lines of text.

    Store:
    - sound file type (MONSTER, BUN)
    - compression type (MP3, OGG, FLAC, None/original e.g. VOC)
    - for each sound:
    -- ID
    -- line_nums of text that play the sound
    -- metadata
    --- type (MONSTER, BUN)
    --- tags
    --- other nodes for BUN data (TBA)
    -- Source (excluding metadata)
    --- type (MONSTER, BUN, external) -> maps to class that can read the file
    --- starting offset
    --- size
    - text comments
    - sound comments
    """

    metadata_marshallers_in = {
        # Convert lip synch data, by splitting CSVs and converting to ints.
        "LipSynch" : lambda ls: map(int, ls.split(', '))
    }
    metadata_marshallers_out = {
        # Convert lip synch data, by convert to CSVs and convertins int to str.
        "LipSynch" : lambda ls: ', '.join(map(str, ls))
    }
    generated_metadata = frozenset([
        "Frequency",
        "Channels",
        "Regions", # we only support 1 region for now
    ])

    bun_group_prefix = "SSNEW"

    def __init__(self):
        """
        text_path = path to the text file containing dialogue.
        resource_type = "MONSTER" or "BUN"
        compression_type = "MP3", "OGG", "FLAC", or "Original"
        sounds = list of SoundEntry objects
        """
        self.text_path = None
        self.resource_type = None
        self.compression_type = None
        self.sounds = []
        self.text_comments = {}
        self.sound_comments = {}

    def __len__(self):
        return len(self.sounds)

    def __str__(self):
        s = (
            ("SpeechMap:\n" +
            "sm> text_path: %s\n" +
            "sm> resource_type: %s\n" +
            "sm> compression_type: %s\n" +
            "sm> sounds: %s\n" +
            "sm> text_comments: %s\n" +
            "sm> sound_comments: %s\n"
            ) %
            (self.text_path, self.resource_type, self.compression_type,
            self.sounds, self.text_comments, self.sound_comments)
        )
        return s

    def _next_sound_id(self):
        if len(self.sounds) == 0:
            return "0"
        if self.resource_type == ssconst.TYPE_MONSTER:
            return max((int(s.id) for s in self.sounds)) + 100 # create an imaginary ID
        elif self.resource_type == ssconst.TYPE_BUN:
            # Allow 999 different sound groups, each with 999 different sounds.

            # Find the next group
            try:
                max_group = max(
                    (
                        int(s.id[5:8])
                        for s
                        in self.sounds
                        if s.id.startswith(self.bun_group_prefix)
                    )
                )
            except ValueError:
                # If no group found, add initial group with new sound ID.
                return self.bun_group_prefix + "001.001"

            # Find the max value in the group. (this should not throw a ValueError,
            #  since we've already established that a group exists from the previous
            #  pass)
            max_id = max(
                (
                    int(s.id[9:])
                    for s
                    in self.sounds
                    if s.id.startswith(self.bun_group_prefix + str(max_group).zfill(3))
                )
            )

            # If value is 999, create a new group.
            if max_id == 999:
                if max_group == 999:
                    raise ssmisc.SSException(
                        "Sorry, you have reached the maximum number of automatically "
                        "generated group name for BUN resources! Please save the "
                        "speech map XML, and manually alter the IDs of your sounds."
                    )
                return self.bun_group_prefix + str(max_group + 1).zfill(3) + ".001"

            # Return something like "SSNEW001.002"
            return self.bun_group_prefix + str(max_group).zfill(3) + "." + str(max_id + 1).zfill(3)


    def _validate_sound(self, sound_path):
        """Raises exception if sound's compression doesn't match the
        speech map's compression."""
        if self.compression_type == "Original":
            raise ssmisc.SSException("Adding sounds that use the original compression format is not supported yet.")
        elif self.compression_type != os.path.splitext(sound_path)[1].upper().lstrip('.'):
            raise ssmisc.SSException("Sound file extension does not appear to match the expected compression format: %s" % self.compression_type)

    def get_resource_file_ext(self):
        if self.resource_type == ssconst.TYPE_MONSTER:
            if self.compression_type == ssconst.COMP_MP3:
                return '.so3'
            elif self.compression_type == ssconst.COMP_OGG:
                return '.sog'
            elif self.compression_type == ssconst.COMP_FLAC:
                return '.sof'
            return '.sou' # default to returning origianl compression
        elif self.resource_type == ssconst.TYPE_BUN:
            return '.bun'
        else:
            raise ssmisc.SSException("Unknown resource type for speech map: %s" % self.resource_type)

    def get_sound_used_by_line(self, line_num):
        """Returns a tuple containing the sound number and the sound object.

        NOTE: while this always returns a list, I don't know of any scenario
        where more than 1 sound will be returned."""
        return [(i, s) for i, s in enumerate(self.sounds) if line_num in s.lines_used]

    def get_aux_data_size_of_sound(self, sound_num):
        """Only used by MONSTER resources. The text that triggers a sound requires
        both the offset to the sound, and the size of the VCTL/VTTL header, which
        contains the aux data (lip synch tags)."""
        if self.resource_type == ssconst.TYPE_MONSTER:
            # 8 for the VCTL/VTTL header, every lip synch tag that follows is 16-bit
            return 8 + (len(self.sounds[sound_num].metadata['LipSynch']) * 2)
        else:
            raise ssmisc.SSException("Unknown resource type for speech map: %s" % self.resource_type)

    def unmap_sound_from_text(self, line_num):
        """Attempts to unmap the line numbers from any associated sounds.
        Throws SSException if no sound is associated with the line of text."""
        sounds = self.get_sound_used_by_line(line_num)
        if len(sounds) == 0:
            raise ssmisc.SSException("No sounds found that map to line #%s" % (line_num,))
        for _, s in sounds:
            s.lines_used.remove(line_num)

    def map_sound_to_text(self, sound_num, line_num):
        sound = self.sounds[sound_num]
        if not line_num in sound.lines_used and \
            len(self.get_sound_used_by_line(line_num)) == 0:
            sound.lines_used.append(line_num)
        else:
            raise ssmisc.SSException(("Could not map sound #%s to line #%s " +
                "because the line is already mapped.") % (sound_num, line_num))

    def _get_default_metadata(self):
        metadata = None
        if self.resource_type == ssconst.TYPE_MONSTER:
            metadata = {
                "LipSynch" : 0xFFF
            }
        elif self.resource_type == ssconst.TYPE_BUN:
           metadata = {
                "Version" : 3,
                "BitDepth" : 16,
                "Frequency" : 22050,
                "Channels" : 1,
                "Regions" : [],
                "Jumps" : [],
                "Syncs" : [],
                "Markers" : []
            }
        else:
            raise ssmisc.SSException("Unsupported resource type: %s" %
                    self.resource_type)
        return metadata

    def _update_bun_metadata(self, metadata, sound_path):
        if self.compression_type == ssconst.COMP_MP3:
            info = mutagen.mp3.MP3(sound_path).info
        elif self.compression_type == ssconst.COMP_OGG:
            info = mutagen.oggvorbis.OggVorbis(sound_path).info
        elif self.compression_type == ssconst.COMP_FLAC:
            info = mutagen.flac.FLAC(sound_path).info
        else:
            raise ssmisc.SSException("Unsupported compression type: %s" %
                    self.compression_type)
        #print info.pprint()
        regions = metadata["Regions"]
        bit_depth = metadata["BitDepth"]
        # Calculate total number of samples in uncompressed sound data.
        region_length = int(math.ceil((info.length * info.sample_rate) *
                         info.channels *
                         (bit_depth / 8.0)))
        #print region_length
        if len(regions) == 0:
            regions.append({"Offset" : 0, "Length" : region_length})
        elif len(regions) == 1:
            regions[0]["Length"] = region_length
        else:
            raise ssmisc.SSException("Unsupported number of regions for sound.")
        # Override some of the the metadata with actual sound's attributes
        metadata["Frequency"] = info.sample_rate
        metadata["Channels"] = info.channels

    def _read_sound(self, new_sound_path):
        # Check that the new sound is in the right format.
        self._validate_sound(new_sound_path)

        # Get the file size
        sound_size = os.stat(new_sound_path).st_size

        # Test for an XML file containing metadata
        metadata = self._get_default_metadata() # Use default metadata if none specified
        xml_path = os.path.splitext(new_sound_path)[0] + ".xml"
        if os.path.isfile(xml_path):
            root = et.parse(xml_path).getroot()
            assert root.tag == "ExternalSound"
            metadata.update(self._load_metadata_from_xml(root.find("SoundEntry")))

        # For BUN files, we need to calculate the length of regions.
        if self.resource_type == ssconst.TYPE_BUN:
            self._update_bun_metadata(metadata, new_sound_path)

        return (sound_size, metadata)

    def add_sound(self, external_sound_path):
        sound_size, metadata = self._read_sound(external_sound_path)

        # Add the sound
        sound = SoundEntry(
            self._next_sound_id(),
            SoundSource(
                external_sound_path,
                0,
                sound_size),
            metadata, # metadata is to be provided
            [] # lines used (none to start with)
        )
        self.sounds.append(sound)

    def remove_sound(self, sound_num):
        del self.sounds[sound_num]

    def replace_sound(self, sound_num, new_sound_path):
        sound_size, metadata = self._read_sound(new_sound_path)

        old_sound = self.sounds[sound_num]
        new_sound = SoundEntry(
            old_sound.id,
            SoundSource(
                new_sound_path,
                0,
                sound_size
            ),
            metadata,
            old_sound.lines_used
        )
        self.sounds[sound_num] = new_sound

    def import_text_sound_map(self, text_sound_map):
        for s in self.sounds:
            try:
                s.lines_used = text_sound_map[s.id]
            except KeyError:
                s.lines_used = []

    def _load_soundentry_from_xml(self, xml_node):
        id = xml_node.find("ID").text
        source_node = xml_node.find("Source")
        if source_node is not None:
            source = SoundSource(source_node.find("Path").text,
                                 int(source_node.find("Offset").text),
                                 int(source_node.find("Size").text))
        else:
            source = None
        metadata = self._get_default_metadata()
        metadata.update(self._load_metadata_from_xml(xml_node))
        lines_used = xml_node.find("LinesUsed").text
        lines_used = [] if lines_used is None else map(int, lines_used.split(', '))
        return (id, source, metadata, lines_used)

    def _load_metadata_from_xml(self, xml_node):
        metadata = {}
        for md in xml_node.find("Metadata").getchildren():
            try:
                data = self.metadata_marshallers_in[md.tag](md.text)
            except KeyError:
                data = eval(md.text) # mega dangerous but I'm mega lazy
            metadata[md.tag] = data
        return metadata

    def load_from_file(self, path):
        root = et.parse(path).getroot()
        assert root.tag == "SpeechMap"
        self.text_path = root.find("TextPath").text
        self.resource_type = root.find("ResourceType").text
        self.compression_type = root.find("CompressionType").text
        self.sounds = []
        for se in root.find("Sounds").findall("SoundEntry"):
            self.sounds.append(SoundEntry(*self._load_soundentry_from_xml(se)))
        self.text_comments = {}
        for tc in root.find("TextComments").findall("Comment"):
            self.text_comments[tc.get('id')] = tc.text
        self.sound_comments = {}
        for sc in root.find("SoundComments").findall("Comment"):
            self.sound_comments[sc.get('id')] = sc.text

    def import_sounds(self, in_path):
        self.sounds = []
        self.sound_comments = {}
        # Get all sound files that match the desired compression type
        file_list = (os.path.join(in_path, f) for f in os.listdir(in_path)
                    if os.path.splitext(f)[1].lower() == ('.' + self.compression_type.lower()))
        for sound_path in file_list:
            # First, read in metadata from XML
            sound_xml_path = os.path.splitext(sound_path)[0] + '.xml'
            id = None
            metadata = self._get_default_metadata()
            lines_used = None
            comment = None
            if os.path.isfile(sound_xml_path):
                root = et.parse(sound_xml_path).getroot()
                assert root.tag == "ExternalSound"
                id, _, input_metadata, lines_used = self._load_soundentry_from_xml(root.find("SoundEntry"))
                metadata.update(input_metadata)
                comment = root.find("Comment").text
            # Use defaults/generated values if no XML found, or values are blank
            if id is None:
                id = self._next_sound_id()
            if lines_used is None:
                lines_used = []

            # For BUN files, we need to calculate the length of regions.
            if self.resource_type == ssconst.TYPE_BUN:
                self._update_bun_metadata(metadata, sound_path)

            # Add the sound
            sound = SoundEntry(
                id,
                SoundSource(
                    sound_path,
                    0,
                    os.stat(sound_path).st_size
                ),
                metadata,
                lines_used
            )
            self.sounds.append(sound)
            # Add the comment
            if comment is not None:
                self.sound_comments[id] = comment
                
        # Sort the sounds by ID, to approximate how they appear in the
        #  original resource files.
        self.sounds.sort(
        key=(lambda s: int(s.id))
            if self.compression_type == ssconst.TYPE_MONSTER
            else (lambda s: s.id)
        )

    def _save_soundentry_to_xml(self, sound, xml_node, source_info=None, omit_generated_metadata=False):
        s_node = et.SubElement(xml_node, "SoundEntry")
        et.SubElement(s_node, "ID").text = sound.id
        # "Source" node is optional
        if source_info != None:
            path, offset, size = source_info
            source_node = et.SubElement(s_node, "Source")
            et.SubElement(source_node, "Path").text = path
            et.SubElement(source_node, "Offset").text = str(offset)
            et.SubElement(source_node, "Size").text = str(size)
        self._save_metadata_to_xml(sound, s_node, omit_generated_metadata=omit_generated_metadata)
        et.SubElement(s_node, "LinesUsed").text = ', '.join(map(str, sound.lines_used))

    def _save_metadata_to_xml(self, sound, xml_node, omit_generated_metadata=False):
        metadata_node = et.SubElement(xml_node, "Metadata")
        for md_k, md_v in sound.metadata.items():
            if omit_generated_metadata and md_k in self.generated_metadata:
                continue
            try:
                data = self.metadata_marshallers_out[md_k](md_v)
            except KeyError:
                data = repr(md_v)
            et.SubElement(metadata_node, md_k).text = data

    def save_to_file(self, path):
        root = et.Element("SpeechMap")

        et.SubElement(root, "TextPath").text = self.text_path
        et.SubElement(root, "ResourceType").text = self.resource_type
        et.SubElement(root, "CompressionType").text = self.compression_type

        sounds_node = et.SubElement(root, "Sounds")
        for s in self.sounds:
            self._save_soundentry_to_xml(s, sounds_node, (s.source.path, s.source.offset, s.source.size))
                
        comments_node = et.SubElement(root, "TextComments")
        for c_k, c_v in self.text_comments.items():
            et.SubElement(comments_node, "Comment", {"id" : c_k}).text = c_v
        comments_node = et.SubElement(root, "SoundComments")
        for c_k, c_v in self.sound_comments.items():
            et.SubElement(comments_node, "Comment", {"id" : c_k}).text = c_v

        ssutil.indent_elementtree(root)
        et.ElementTree(root).write(path)

    def export_sounds(self, out_path, text_data):
        """Accepts an output directory, and a TextData object (to generate file names)"""
        buffer_size = 102400 # 100 KB
        unused_counter = 0
        pad = 4
        for s in self.sounds:
            source = s.source

            if len(s.lines_used) > 0:
                line_id = s.lines_used[0] # just grab the first one, other lines we handle later.
                text_line = text_data[line_id]
                if len(text_line) > 25: # limit filename length
                    text_line = text_line[:25]
                text_line = ''.join(c for c in text_line if c in ssconst.setValidChars)
                text_line = text_line.rstrip('. ')
                file_prefix = str(line_id).zfill(pad) + "_" + text_line
            else:
                file_prefix = "unused_" + str(unused_counter).zfill(pad)
                unused_counter += 1

            file_ext = '.' + self.compression_type.lower()
            output_file_name = file_prefix + file_ext

            output_file_path = os.path.join(out_path, output_file_name)
            # Do a buffered copy of the sound file
            with file(source.path, 'rb') as source_file:
                with file(output_file_path, 'wb') as output_file:
                    left_to_read = source.size
                    source_file.seek(source.offset)
                    while left_to_read > 0:
                        read_size = min(left_to_read, buffer_size)
                        output_file.write(source_file.read(read_size))
                        left_to_read -= read_size

            # Write extra information to XML file
            root = et.Element("ExternalSound")
            self._save_soundentry_to_xml(s, root, omit_generated_metadata=True)

            # Output a comment if it exists
            comment = et.SubElement(root, "Comment")
            if self.sound_comments.has_key(s.id):
                comment.text = self.sound_comments[s.id]

            ssutil.indent_elementtree(root)
            et.ElementTree(root).write(os.path.join(out_path, file_prefix + '.xml'))
